@@ -36,9 +36,14 @@ static avifBool avifPrepareReformatState(const avifImage * image, const avifRGBI
         return AVIF_FALSE;
     }
 
+    if ((image->matrixCoefficients == AVIF_MATRIX_COEFFICIENTS_IDENTITY) && (image->yuvFormat != AVIF_PIXEL_FORMAT_YUV444)) {
+        return AVIF_FALSE;
+    }
+
     if (image->yuvFormat == AVIF_PIXEL_FORMAT_NONE) {
         return AVIF_FALSE;
     }
+
     avifGetPixelFormatInfo(image->yuvFormat, &state->formatInfo);
     avifCalcYUVCoefficients(image, &state->kr, &state->kg, &state->kb);
     state->mode = AVIF_REFORMAT_MODE_YUV_COEFFICIENTS;
@@ -109,8 +114,10 @@ static avifBool avifPrepareReformatState(const avifImage * image, const avifRGBI
     state->rgbMaxChannelF = (float)state->rgbMaxChannel;
     state->biasY = (state->yuvRange == AVIF_RANGE_LIMITED) ? (float)(16 << (state->yuvDepth - 8)) : 0.0f;
     state->biasUV = (float)(1 << (state->yuvDepth - 1));
+    state->biasA = (image->alphaRange == AVIF_RANGE_LIMITED) ? (float)(16 << (state->yuvDepth - 8)) : 0.0f;
     state->rangeY = (float)((state->yuvRange == AVIF_RANGE_LIMITED) ? (219 << (state->yuvDepth - 8)) : state->yuvMaxChannel);
     state->rangeUV = (float)((state->yuvRange == AVIF_RANGE_LIMITED) ? (224 << (state->yuvDepth - 8)) : state->yuvMaxChannel);
+    state->rangeA = (float)((image->alphaRange == AVIF_RANGE_LIMITED) ? (219 << (state->yuvDepth - 8)) : state->yuvMaxChannel);
 
     uint32_t cpCount = 1 << image->depth;
     if (state->mode == AVIF_REFORMAT_MODE_IDENTITY) {
@@ -126,6 +133,23 @@ static avifBool avifPrepareReformatState(const avifImage * image, const avifRGBI
         }
     }
 
+    state->toRGBAlphaMode = AVIF_ALPHA_MULTIPLY_MODE_NO_OP;
+    if (image->alphaPlane) {
+        if (!avifRGBFormatHasAlpha(rgb->format) || rgb->ignoreAlpha) {
+            // if we are converting some image with alpha into a format without alpha, we should do 'premultiply alpha' before
+            // discarding alpha plane. This has the same effect of rendering this image on a black background, which makes sense.
+            if (!image->alphaPremultiplied) {
+                state->toRGBAlphaMode = AVIF_ALPHA_MULTIPLY_MODE_MULTIPLY;
+            }
+        } else {
+            if (!image->alphaPremultiplied && rgb->alphaPremultiplied) {
+                state->toRGBAlphaMode = AVIF_ALPHA_MULTIPLY_MODE_MULTIPLY;
+            } else if (image->alphaPremultiplied && !rgb->alphaPremultiplied) {
+                state->toRGBAlphaMode = AVIF_ALPHA_MULTIPLY_MODE_UNMULTIPLY;
+            }
+        }
+    }
+
     return AVIF_TRUE;
 }
 
@@ -164,9 +188,15 @@ avifResult avifImageRGBToYUV(avifImage * image, const avifRGBImage * rgb)
         return AVIF_RESULT_REFORMAT_FAILED;
     }
 
+    avifAlphaMultiplyMode alphaMode = AVIF_ALPHA_MULTIPLY_MODE_NO_OP;
     avifImageAllocatePlanes(image, AVIF_PLANES_YUV);
     if (avifRGBFormatHasAlpha(rgb->format) && !rgb->ignoreAlpha) {
         avifImageAllocatePlanes(image, AVIF_PLANES_A);
+        if (!rgb->alphaPremultiplied && image->alphaPremultiplied) {
+            alphaMode = AVIF_ALPHA_MULTIPLY_MODE_MULTIPLY;
+        } else if (rgb->alphaPremultiplied && !image->alphaPremultiplied) {
+            alphaMode = AVIF_ALPHA_MULTIPLY_MODE_UNMULTIPLY;
+        }
     }
 
     const float kr = state.kr;
@@ -211,8 +241,7 @@ avifResult avifImageRGBToYUV(avifImage * image, const avifRGBImage * rgb)
                         rgbPixel[2] = rgb->pixels[state.rgbOffsetBytesB + (i * state.rgbPixelBytes) + (j * rgb->rowBytes)] / rgbMaxChannelF;
                     }
 
-                    if (avifRGBFormatHasAlpha(rgb->format) && !rgb->ignoreAlpha &&
-                        (rgb->alphaPremultiplied != image->alphaPremultiplied)) {
+                    if (alphaMode != AVIF_ALPHA_MULTIPLY_MODE_NO_OP) {
                         float a;
                         if (state.rgbChannelBytes > 1) {
                             a = *((uint16_t *)(&rgb->pixels[state.rgbOffsetBytesA + (i * state.rgbPixelBytes) + (j * rgb->rowBytes)])) /
@@ -221,8 +250,7 @@ avifResult avifImageRGBToYUV(avifImage * image, const avifRGBImage * rgb)
                             a = rgb->pixels[state.rgbOffsetBytesA + (i * state.rgbPixelBytes) + (j * rgb->rowBytes)] / rgbMaxChannelF;
                         }
 
-                        if (image->alphaPremultiplied) {
-                            // multiply
+                        if (alphaMode == AVIF_ALPHA_MULTIPLY_MODE_MULTIPLY) {
                             if (a == 0) {
                                 rgbPixel[0] = 0;
                                 rgbPixel[1] = 0;
@@ -233,7 +261,7 @@ avifResult avifImageRGBToYUV(avifImage * image, const avifRGBImage * rgb)
                                 rgbPixel[2] *= a;
                             }
                         } else {
-                            // unmultiply
+                            // alphaMode == AVIF_ALPHA_MULTIPLY_MODE_UNMULTIPLY
                             if (a == 0) {
                                 rgbPixel[0] = 0;
                                 rgbPixel[1] = 0;
@@ -381,6 +409,7 @@ avifResult avifImageRGBToYUV(avifImage * image, const avifRGBImage * rgb)
     return AVIF_RESULT_OK;
 }
 
+// Note: This function handles alpha (un)multiply.
 static avifResult avifImageYUVAnyToRGBAnySlow(const avifImage * image,
                                               avifRGBImage * rgb,
                                               avifReformatState * state,
@@ -399,9 +428,11 @@ static avifResult avifImageYUVAnyToRGBAnySlow(const avifImage * image,
     const uint8_t * yPlane = image->yuvPlanes[AVIF_CHAN_Y];
     const uint8_t * uPlane = image->yuvPlanes[AVIF_CHAN_U];
     const uint8_t * vPlane = image->yuvPlanes[AVIF_CHAN_V];
+    const uint8_t * aPlane = image->alphaPlane;
     const uint32_t yRowBytes = image->yuvRowBytes[AVIF_CHAN_Y];
     const uint32_t uRowBytes = image->yuvRowBytes[AVIF_CHAN_U];
     const uint32_t vRowBytes = image->yuvRowBytes[AVIF_CHAN_V];
+    const uint32_t aRowBytes = image->alphaRowBytes;
 
     // Various observations and limits
     const avifBool hasColor = (uPlane && vPlane && (image->yuvFormat != AVIF_PIXEL_FORMAT_YUV400));
@@ -414,11 +445,13 @@ static avifResult avifImageYUVAnyToRGBAnySlow(const avifImage * image,
     for (uint32_t j = 0; j < image->height; ++j) {
         const uint32_t uvJ = j >> state->formatInfo.chromaShiftY;
         const uint8_t * ptrY8 = &yPlane[j * yRowBytes];
-        const uint8_t * ptrU8 = &uPlane[(uvJ * uRowBytes)];
-        const uint8_t * ptrV8 = &vPlane[(uvJ * vRowBytes)];
+        const uint8_t * ptrU8 = uPlane ? &uPlane[(uvJ * uRowBytes)] : NULL;
+        const uint8_t * ptrV8 = vPlane ? &vPlane[(uvJ * vRowBytes)] : NULL;
+        const uint8_t * ptrA8 = aPlane ? &aPlane[j * aRowBytes] : NULL;
         const uint16_t * ptrY16 = (const uint16_t *)ptrY8;
         const uint16_t * ptrU16 = (const uint16_t *)ptrU8;
         const uint16_t * ptrV16 = (const uint16_t *)ptrV8;
+        const uint16_t * ptrA16 = (const uint16_t *)ptrA8;
 
         uint8_t * ptrR = &rgb->pixels[state->rgbOffsetBytesR + (j * rgb->rowBytes)];
         uint8_t * ptrG = &rgb->pixels[state->rgbOffsetBytesG + (j * rgb->rowBytes)];
@@ -588,9 +621,47 @@ static avifResult avifImageYUVAnyToRGBAnySlow(const avifImage * image,
                 B = Y;
             }
 
-            const float Rc = AVIF_CLAMP(R, 0.0f, 1.0f);
-            const float Gc = AVIF_CLAMP(G, 0.0f, 1.0f);
-            const float Bc = AVIF_CLAMP(B, 0.0f, 1.0f);
+            float Rc = AVIF_CLAMP(R, 0.0f, 1.0f);
+            float Gc = AVIF_CLAMP(G, 0.0f, 1.0f);
+            float Bc = AVIF_CLAMP(B, 0.0f, 1.0f);
+
+            if (state->toRGBAlphaMode != AVIF_ALPHA_MULTIPLY_MODE_NO_OP) {
+                // Calculate A
+                uint16_t unormA;
+                if (image->depth == 8) {
+                    unormA = ptrA8[i];
+                } else {
+                    unormA = AVIF_MIN(ptrA16[i], yuvMaxChannel);
+                }
+                const float A = (unormA - state->biasA) / state->rangeA;
+                const float Ac = AVIF_CLAMP(A, 0.0f, 1.0f);
+
+                if (state->toRGBAlphaMode == AVIF_ALPHA_MULTIPLY_MODE_MULTIPLY) {
+                    if (Ac == 0.0f) {
+                        Rc = 0.0f;
+                        Gc = 0.0f;
+                        Bc = 0.0f;
+                    } else if (Ac < 1.0f) {
+                        Rc *= Ac;
+                        Gc *= Ac;
+                        Bc *= Ac;
+                    }
+                } else {
+                    // state->toRGBAlphaMode == AVIF_ALPHA_MULTIPLY_MODE_UNMULTIPLY
+                    if (Ac == 0.0f) {
+                        Rc = 0.0f;
+                        Gc = 0.0f;
+                        Bc = 0.0f;
+                    } else if (Ac < 1.0f) {
+                        Rc /= Ac;
+                        Gc /= Ac;
+                        Bc /= Ac;
+                        Rc = AVIF_MIN(Rc, 1.0f);
+                        Gc = AVIF_MIN(Gc, 1.0f);
+                        Bc = AVIF_MIN(Bc, 1.0f);
+                    }
+                }
+            }
 
             if (rgb->depth == 8) {
                 *ptrR = (uint8_t)(0.5f + (Rc * rgbMaxChannelF));
@@ -1008,17 +1079,21 @@ avifResult avifImageYUVToRGB(const avifImage * image, avifRGBImage * rgb)
         return AVIF_RESULT_REFORMAT_FAILED;
     }
 
+    avifAlphaMultiplyMode alphaMultiplyMode = state.toRGBAlphaMode;
     avifBool convertedWithLibYUV = AVIF_FALSE;
-    avifResult libyuvResult = avifImageYUVToRGBLibYUV(image, rgb);
-    if (libyuvResult == AVIF_RESULT_OK) {
-        convertedWithLibYUV = AVIF_TRUE;
-    } else {
-        if (libyuvResult != AVIF_RESULT_NOT_IMPLEMENTED) {
-            return libyuvResult;
+    if (alphaMultiplyMode == AVIF_ALPHA_MULTIPLY_MODE_NO_OP || avifRGBFormatHasAlpha(rgb->format)) {
+        avifResult libyuvResult = avifImageYUVToRGBLibYUV(image, rgb);
+        if (libyuvResult == AVIF_RESULT_OK) {
+            convertedWithLibYUV = AVIF_TRUE;
+        } else {
+            if (libyuvResult != AVIF_RESULT_NOT_IMPLEMENTED) {
+                return libyuvResult;
+            }
         }
     }
 
-    if (avifRGBFormatHasAlpha(rgb->format) && !rgb->ignoreAlpha) {
+    // Reformat alpha, if user asks for it, or (un)multiply processing needs it.
+    if (avifRGBFormatHasAlpha(rgb->format) && (!rgb->ignoreAlpha || (alphaMultiplyMode != AVIF_ALPHA_MULTIPLY_MODE_NO_OP))) {
         avifAlphaParams params;
 
         params.width = rgb->width;
@@ -1046,105 +1121,113 @@ avifResult avifImageYUVToRGB(const avifImage * image, avifRGBImage * rgb)
         }
     }
 
-    // Do this after alpha conversion
-    if (convertedWithLibYUV) {
-        return AVIF_RESULT_OK;
-    }
+    if (!convertedWithLibYUV) {
+        // libyuv is either unavailable or unable to perform the specific conversion required here.
+        // Look over the available built-in "fast" routines for YUV->RGB conversion and see if one
+        // fits the current combination, or as a last resort, call avifImageYUVAnyToRGBAnySlow(),
+        // which handles every possibly YUV->RGB combination, but very slowly (in comparison).
 
-    avifChromaUpsampling chromaUpsampling;
-    switch (rgb->chromaUpsampling) {
-        case AVIF_CHROMA_UPSAMPLING_AUTOMATIC:
-        case AVIF_CHROMA_UPSAMPLING_BEST_QUALITY:
-        case AVIF_CHROMA_UPSAMPLING_BILINEAR:
-        default:
-            chromaUpsampling = AVIF_CHROMA_UPSAMPLING_BILINEAR;
-            break;
+        avifResult convertResult = AVIF_RESULT_NOT_IMPLEMENTED;
 
-        case AVIF_CHROMA_UPSAMPLING_FASTEST:
-        case AVIF_CHROMA_UPSAMPLING_NEAREST:
-            chromaUpsampling = AVIF_CHROMA_UPSAMPLING_NEAREST;
-            break;
-    }
+        avifChromaUpsampling chromaUpsampling;
+        switch (rgb->chromaUpsampling) {
+            case AVIF_CHROMA_UPSAMPLING_AUTOMATIC:
+            case AVIF_CHROMA_UPSAMPLING_BEST_QUALITY:
+            case AVIF_CHROMA_UPSAMPLING_BILINEAR:
+            default:
+                chromaUpsampling = AVIF_CHROMA_UPSAMPLING_BILINEAR;
+                break;
 
-    const avifBool hasColor =
-        (image->yuvRowBytes[AVIF_CHAN_U] && image->yuvRowBytes[AVIF_CHAN_V] && (image->yuvFormat != AVIF_PIXEL_FORMAT_YUV400));
+            case AVIF_CHROMA_UPSAMPLING_FASTEST:
+            case AVIF_CHROMA_UPSAMPLING_NEAREST:
+                chromaUpsampling = AVIF_CHROMA_UPSAMPLING_NEAREST;
+                break;
+        }
 
-    avifResult convertResult = AVIF_RESULT_NOT_IMPLEMENTED;
+        const avifBool hasColor =
+            (image->yuvRowBytes[AVIF_CHAN_U] && image->yuvRowBytes[AVIF_CHAN_V] && (image->yuvFormat != AVIF_PIXEL_FORMAT_YUV400));
 
-    if (!hasColor || (image->yuvFormat == AVIF_PIXEL_FORMAT_YUV444) || (chromaUpsampling == AVIF_CHROMA_UPSAMPLING_NEAREST)) {
-        // None of these fast paths currently support bilinear upsampling, so avoid all of them
-        // unless the YUV data isn't subsampled or they explicitly requested AVIF_CHROMA_UPSAMPLING_NEAREST.
+        if ((!hasColor || (image->yuvFormat == AVIF_PIXEL_FORMAT_YUV444) || (chromaUpsampling == AVIF_CHROMA_UPSAMPLING_NEAREST)) &&
+            (alphaMultiplyMode == AVIF_ALPHA_MULTIPLY_MODE_NO_OP || avifRGBFormatHasAlpha(rgb->format))) {
+            // Explanations on the above conditional:
+            // * None of these fast paths currently support bilinear upsampling, so avoid all of them
+            //   unless the YUV data isn't subsampled or they explicitly requested AVIF_CHROMA_UPSAMPLING_NEAREST.
+            // * None of these fast paths currently handle alpha (un)multiply, so avoid all of them
+            //   if we can't do alpha (un)multiply as a separated post step (destination format doesn't have alpha).
 
-        if (state.mode == AVIF_REFORMAT_MODE_IDENTITY) {
-            if ((image->depth == 8) && (rgb->depth == 8) && (image->yuvFormat == AVIF_PIXEL_FORMAT_YUV444) &&
-                (image->yuvRange == AVIF_RANGE_FULL)) {
-                convertResult = avifImageIdentity8ToRGB8ColorFullRange(image, rgb, &state);
-            }
-
-            // TODO: Add more fast paths for identity
-        } else if (state.mode == AVIF_REFORMAT_MODE_YUV_COEFFICIENTS) {
-            if (image->depth > 8) {
-                // yuv:u16
-
-                if (rgb->depth > 8) {
-                    // yuv:u16, rgb:u16
-
-                    if (hasColor) {
-                        convertResult = avifImageYUV16ToRGB16Color(image, rgb, &state);
-                    } else {
-                        convertResult = avifImageYUV16ToRGB16Mono(image, rgb, &state);
-                    }
-                } else {
-                    // yuv:u16, rgb:u8
-
-                    if (hasColor) {
-                        convertResult = avifImageYUV16ToRGB8Color(image, rgb, &state);
-                    } else {
-                        convertResult = avifImageYUV16ToRGB8Mono(image, rgb, &state);
-                    }
+            if (state.mode == AVIF_REFORMAT_MODE_IDENTITY) {
+                if ((image->depth == 8) && (rgb->depth == 8) && (image->yuvFormat == AVIF_PIXEL_FORMAT_YUV444) &&
+                    (image->yuvRange == AVIF_RANGE_FULL)) {
+                    convertResult = avifImageIdentity8ToRGB8ColorFullRange(image, rgb, &state);
                 }
-            } else {
-                // yuv:u8
 
-                if (rgb->depth > 8) {
-                    // yuv:u8, rgb:u16
+                // TODO: Add more fast paths for identity
+            } else if (state.mode == AVIF_REFORMAT_MODE_YUV_COEFFICIENTS) {
+                if (image->depth > 8) {
+                    // yuv:u16
 
-                    if (hasColor) {
-                        convertResult = avifImageYUV8ToRGB16Color(image, rgb, &state);
+                    if (rgb->depth > 8) {
+                        // yuv:u16, rgb:u16
+
+                        if (hasColor) {
+                            convertResult = avifImageYUV16ToRGB16Color(image, rgb, &state);
+                        } else {
+                            convertResult = avifImageYUV16ToRGB16Mono(image, rgb, &state);
+                        }
                     } else {
-                        convertResult = avifImageYUV8ToRGB16Mono(image, rgb, &state);
+                        // yuv:u16, rgb:u8
+
+                        if (hasColor) {
+                            convertResult = avifImageYUV16ToRGB8Color(image, rgb, &state);
+                        } else {
+                            convertResult = avifImageYUV16ToRGB8Mono(image, rgb, &state);
+                        }
                     }
                 } else {
-                    // yuv:u8, rgb:u8
+                    // yuv:u8
 
-                    if (hasColor) {
-                        convertResult = avifImageYUV8ToRGB8Color(image, rgb, &state);
+                    if (rgb->depth > 8) {
+                        // yuv:u8, rgb:u16
+
+                        if (hasColor) {
+                            convertResult = avifImageYUV8ToRGB16Color(image, rgb, &state);
+                        } else {
+                            convertResult = avifImageYUV8ToRGB16Mono(image, rgb, &state);
+                        }
                     } else {
-                        convertResult = avifImageYUV8ToRGB8Mono(image, rgb, &state);
+                        // yuv:u8, rgb:u8
+
+                        if (hasColor) {
+                            convertResult = avifImageYUV8ToRGB8Color(image, rgb, &state);
+                        } else {
+                            convertResult = avifImageYUV8ToRGB8Mono(image, rgb, &state);
+                        }
                     }
                 }
             }
         }
-    }
 
-    if (convertResult == AVIF_RESULT_NOT_IMPLEMENTED) {
-        // If we get here, there is no fast path for this combination. Time to be slow!
-        convertResult = avifImageYUVAnyToRGBAnySlow(image, rgb, &state, chromaUpsampling);
-    }
+        if (convertResult == AVIF_RESULT_NOT_IMPLEMENTED) {
+            // If we get here, there is no fast path for this combination. Time to be slow!
+            convertResult = avifImageYUVAnyToRGBAnySlow(image, rgb, &state, chromaUpsampling);
 
-    if (convertResult != AVIF_RESULT_OK) {
-        return convertResult;
-    }
+            // The slow path also handles alpha (un)multiply, so forget the operation here.
+            alphaMultiplyMode = AVIF_ALPHA_MULTIPLY_MODE_NO_OP;
+        }
 
-    if (avifRGBFormatHasAlpha(rgb->format) && !rgb->ignoreAlpha) {
-        if (image->alphaPremultiplied && !rgb->alphaPremultiplied) {
-            convertResult = avifRGBImageUnpremultiplyAlpha(rgb);
-        } else if (!image->alphaPremultiplied && rgb->alphaPremultiplied) {
-            convertResult = avifRGBImagePremultiplyAlpha(rgb);
+        if (convertResult != AVIF_RESULT_OK) {
+            return convertResult;
         }
     }
 
-    return convertResult;
+    // Process alpha premultiplication, if necessary
+    if (alphaMultiplyMode == AVIF_ALPHA_MULTIPLY_MODE_MULTIPLY) {
+        return avifRGBImagePremultiplyAlpha(rgb);
+    } else if (alphaMultiplyMode == AVIF_ALPHA_MULTIPLY_MODE_UNMULTIPLY) {
+        return avifRGBImageUnpremultiplyAlpha(rgb);
+    }
+
+    return AVIF_RESULT_OK;
 }
 
 // Limited -> Full
